from emspy.query.ems import EMS
from emspy.query.fleet import Fleet
from emspy.query.aircraft import Aircraft
from emspy.query.airport import Airport
from emspy.query.flightphase import FlightPhase
from emspy.query.localdata import LocalData
from emspy.query.flight import Flight
from emspy.query.analytic import Analytic
from emspy.query.fltquery import FltQuery
from emspy.query.tsquery import TSeriesQuery
from emspy.query.profile import Profile
from emspy.query.dbmodify import InsertQuery

